/*-
 * Copyright (C)2010..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 * Used tcpmssd (c) 2000 by Ruslan Ermilov as reference.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2010..2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: mssseparated.c,v 1.19 2024/04/13 13:11:17 babolo Exp $"

#define BLIN_COMPAT   4
#define Bpars_COMPAT  4
#define RECOBE_COMPAT 6
#define MIFE_COMPAT   5

#include <sys/types.h>
#include <sys/event.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sysexits.h>
#include <strings.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include "mssseparated.h"

static char sterm, shup;

static void
sigterm_handler(int sig) {
    (void)sig;
    sterm = 1;
}

static void
sighup_handler(int sig) {
    (void)sig;
    shup = 1;
}

__attribute__((__noreturn__)) static void
usage(int ex) {
    fprintf( stderr
           , "mssseparated @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "   Separates TCP SYN packets on ipfw divert socket by MSS\n"
             "\n"
             "Usage:\n"
             "mssseparated [-flags]\n"
             "   -D         daemonize\n"
             "   -d port    divert port\n"
             "   -E -fl env use environment variable env as argument of -fl\n"
             "   -e         use envirinment on config file\n"
             "   -F         default path to config file ("MSSSEPARATE_FCONF")\n"
             "   -f file    path to config file\n"
             "   -L         default path to log file ("MSSSEPARATE_FLOG")\n"
             "   -l file    path to log file\n"
             "   -P         default path to pid file ("MSSSEPARATE_FPID")\n"
             "   -p         path to pid file\n"
             "   -q         quite\n"
             "   -s sec     timeout on resume\n"
             "   -v[v[v..]] more verbose\n"
             "   -?         this text\n"
             "   -=0        blin_ctl(BLIN_CTL_DUMP)\n"
             "   -=1        babolo_optest()\n"
             "   -=2        babolo_optext()\n"
             "   -=3        babolo_dumpopts()\n"
             "   -=C        printout config\n"
             "   -=O        blin_ctl(BLIN_CTL_DUMP) after config\n"
           )
    ;
    exit(ex);
}

static void
pidsave(mssseparated_icfg const *icfg) {
#   define blin_internal_flags (icfg->flags)
    if  (icfg->pidfi) {
        FILE *pidfile;

        if  (!(pidfile = fopen(icfg->pidfi, "w"))) {
            ifBLIN_QW0("fopen %s", icfg->pidfi);
        }
        fprintf(pidfile, "%d\n-d %u\n", (int)getpid(), (u_int)icfg->port);
        fclose(pidfile);
        ifBLIN_QX1("PID %d -d %u", (int)getpid(), (u_int)icfg->port);
    }
#   undef blin_internal_flags
}

#define REEND(E)                            \
    if  (icfg->flags & MSSSEPARATE_SLEEP) { \
        sleep(icfg->seconds);               \
        ifBLIN_QW0("Resume + Restart");     \
        goto in;                            \
    } else {                                \
        ex = E;                             \
        goto out;                           \
    }

static void
printlog( mssseparated_icfg const *icfg
        , struct ip               *ipacket
        , tcpshdr                 *tcpacket
        , int32_t                  mss
        , in_port_t                port
        , ssize_t                  plen
        , int                      mode
        ) {
#   define blin_internal_flags (icfg->flags)
    struct protoent *proto;
    char const      *ops = "";

    proto = getprotobynumber(ipacket->ip_p);
    switch(mode) {
    case 0:
        if  (!tcpacket) {
            ifBLIN_QX0("No TCP header at mode 0");
            ops = " Oops!!";
            goto oops;
        }
        ifBLIN_QX1( "ip%u %s %u.%u.%u.%u:%u->%u.%u.%u.%u:%u mss=%d%s%s%s%s%s%s%s%s rule=%u"
                  , ipacket->ip_v
                  , proto->p_name
                  , ipacket->ip_src.s_addr & 0x0FF
                  , (ipacket->ip_src.s_addr >> 8) & 0x0FF
                  , (ipacket->ip_src.s_addr >> 16) & 0x0FF
                  , (ipacket->ip_src.s_addr >> 24) & 0x0FF
                  , ntohs(tcpacket->th_sport)
                  , ipacket->ip_dst.s_addr & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 8) & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 16) & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 24) & 0x0FF
                  , ntohs(tcpacket->th_dport)
                  , mss
                  , (tcpacket->th_flags & TH_FIN)  ? " fin"  : ""
                  , (tcpacket->th_flags & TH_SYN)  ? " syn"  : ""
                  , (tcpacket->th_flags & TH_RST)  ? " rst"  : ""
                  , (tcpacket->th_flags & TH_PUSH) ? " push" : ""
                  , (tcpacket->th_flags & TH_ACK)  ? " ack"  : ""
                  , (tcpacket->th_flags & TH_URG)  ? " urg"  : ""
                  , (tcpacket->th_flags & TH_ECE)  ? " ece"  : ""
                  , (tcpacket->th_flags & TH_CWR)  ? " cwr"  : ""
                  , port
                  );
        break;
    case 1:
        if  (!tcpacket) {
            ifBLIN_QX0("No TCP header at mode 1");
            ops = " Oops!";
            goto oops;
        }
        ifBLIN_QX1( "ip%u %s %u.%u.%u.%u:%u->%u.%u.%u.%u:%u%s%s%s%s%s%s%s%s p%zd h%u o%u rule=%u"
                  , ipacket->ip_v
                  , proto->p_name
                  , ipacket->ip_src.s_addr & 0x0FF
                  , (ipacket->ip_src.s_addr >> 8) & 0x0FF
                  , (ipacket->ip_src.s_addr >> 16) & 0x0FF
                  , (ipacket->ip_src.s_addr >> 24) & 0x0FF
                  , ntohs(tcpacket->th_sport)
                  , ipacket->ip_dst.s_addr & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 8) & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 16) & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 24) & 0x0FF
                  , ntohs(tcpacket->th_dport)
                  , (tcpacket->th_flags & TH_FIN)  ? " fin"  : ""
                  , (tcpacket->th_flags & TH_SYN)  ? " syn"  : ""
                  , (tcpacket->th_flags & TH_RST)  ? " rst"  : ""
                  , (tcpacket->th_flags & TH_PUSH) ? " push" : ""
                  , (tcpacket->th_flags & TH_ACK)  ? " ack"  : ""
                  , (tcpacket->th_flags & TH_URG)  ? " urg"  : ""
                  , (tcpacket->th_flags & TH_ECE)  ? " ece"  : ""
                  , (tcpacket->th_flags & TH_CWR)  ? " cwr"  : ""
                  , plen
                  , ipacket->ip_hl << 2
                  , tcpacket->th_off << 2
                  , port
                  );
        break;
    default:
oops:
        ifBLIN_QX1( "ip%u %s %u.%u.%u.%u->%u.%u.%u.%u p%zd h%u rule=%u%s"
                  , ipacket->ip_v
                  , proto->p_name
                  , ipacket->ip_src.s_addr & 0x0FF
                  , (ipacket->ip_src.s_addr >> 8) & 0x0FF
                  , (ipacket->ip_src.s_addr >> 16) & 0x0FF
                  , (ipacket->ip_src.s_addr >> 24) & 0x0FF
                  , ipacket->ip_dst.s_addr & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 8) & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 16) & 0x0FF
                  , (ipacket->ip_dst.s_addr >> 24) & 0x0FF
                  , plen
                  , ipacket->ip_hl << 2
                  , port
                  , ops
                  );
    }
    return;
#   undef blin_internal_flags
}

int
main(int argc, char **argv) {
    struct ip          *ipacket = NULL;
    char const         *flar    = "Dd:eFf:Ll:Pp:qs:v?=:";
    mssseparated_icfg  *icfg    ;
    struct sockaddr_in  sin     ;
    struct kevent       kev     ;
    babolo_opts        *bos     ;
    int                 ex      ;
    int                 kq      ; /* kqueue  */
    int                 s       ; /* DIVERT  */

    icfg = malloc(MSSSEPARATE_ILEN * sizeof(mssseparated_icfgentry) + sizeof(mssseparated_icfg));
    if  (!icfg) {
        ifBLIN_QW0("icfg");
        ex = EX_SOFTWARE;
        goto out;
    }
    blin_ctl(BLIN_CTL_FLAG | BLIN_CTL_FSET, (icfg->flags = BLIN_GEN0));
#   define blin_internal_flags (icfg->flags)
in:
    icfg->ilen = MSSSEPARATE_ILEN;
    icfg->top = 0;
    icfg->pidfi = NULL;
    icfg->confi = NULL;
    icfg->logfi = NULL;
    icfg->logg = NULL;
    icfg->port = 0;
    s = -1;
    icfg->port = 0;
    ex = EX_OK;
    icfg->seconds = 1;
    icfg->pidfi = NULL;
    if  (!(bos = babolo_openopts(BLIN_GEN0 | Bpars_NONU | Bpars_SYME | Bpars_ESYM, 'Z'))) {
        ifBLIN_QW0("babolo_openopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("babolo_setopts");
        ex = EX_SOFTWARE;
        goto out;
    }
    for (int c; 0 < (c = babolo_getopts(bos));) {
        switch (c) {
        case 'D': icfg->flags |= MSSSEPARATE_DAEMON;
                ; break;
        case 'd': icfg->port = (u_int16_t)babolo_getoptunum(bos, 0);
                ; if  (!!errno) {
                      ifBLIN_QW0("port N absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'e': icfg->flags |= MSSSEPARATE_CFGENV;
                ; break;
        case 'F': icfg->confi = MSSSEPARATE_FCONF;
                ; break;
        case 'f': icfg->confi = babolo_getoptsarg(bos);
                ; if  (!icfg->confi) {
                      ifBLIN_QX0("conf file name absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'L': icfg->logfi = MSSSEPARATE_FLOG;
                ; break;
        case 'l': icfg->logfi = babolo_getoptsarg(bos);
                ; if  (!icfg->logfi) {
                      ifBLIN_QX0("log file name absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'P': icfg->pidfi = MSSSEPARATE_FPID;
                ; break;
        case 'p': icfg->pidfi = babolo_getoptsarg(bos);
                ; if  (!icfg->pidfi) {
                      ifBLIN_QX0("PID file name absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; break;
        case 'q': BLIN_QUIET(icfg->flags);
                ; break;
        case 's': icfg->seconds = (unsigned int)babolo_getoptunum(bos, 0);
                ; if  (!!errno) {
                      ifBLIN_QX0("sleep time absent");
                      ex = EX_USAGE;
                      goto out;
                  }
                ; icfg->flags |= MSSSEPARATE_SLEEP;
                ; break;
        case 'v': BLIN_VERBOSE(icfg->flags);
                ; break;
        case '?': usage(EX_OK);
        case '=': {   const u_char *Tflag;

                ;     if  (!(Tflag = (const u_char *)babolo_getoptsarg(bos))) {
                ;         ifBLIN_QW0("babolo_getopts");
                ;         ex = EX_USAGE;
                ;         goto out;
                ;     }
                ;     ifBLIN_QX4("%02X(%c)", *Tflag & 0x0FF, *Tflag);
                ;     switch(*Tflag) {
                ;     case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                ;     case '1': babolo_optest(flar);               exit(EX_OK);
                ;     case '2': babolo_optext(bos);                exit(EX_OK);
                ;     case '3': babolo_dumpopts(bos);              exit(EX_OK);
                ;     case 'C': icfg->flags |= MSSSEPARATE_OUCONF; break;
                ;     case 'O': icfg->flags |= MSSSEPARATE_CTLDMP; break;
                ;     default : usage(EX_USAGE);
                ; }   }
                ; break;
        default : usage(EX_USAGE);
        }
        if  (ex) goto out;
    }
    if  (!!icfg->logfi) {
        blin_ctl(BLIN_CTL_FNAM | BLIN_GEN7, icfg->logfi);
        blin_ctl(BLIN_CTL_WHAT | BLIN_CTL_FEQU | BLIN_GEN7, BLIN_MODEOAU | BLIN_MODPRNT);
    } else if (MSSSEPARATE_DAEMON & icfg->flags) {
        blin_ctl(BLIN_CTL_WHAT | BLIN_CTL_FEQU | BLIN_GEN7, BLIN_MODEOAU | BLIN_MODLOGR);
    }
    if  (!(icfg = mssseparated_readcfg(icfg))) {
        ex = EX_OSERR;
        goto out;
    }
    if  (MSSSEPARATE_OUCONF & icfg->flags) mssseparated_printcfg(icfg);
    if  ((MSSSEPARATE_DAEMON & icfg->flags) && daemon(0, 0)) {
        ifBLIN_QW0("daemon()");
        ex = EX_OSERR;
        goto out;
    }
    if  (MSSSEPARATE_CTLDMP & icfg->flags) blin_ctl(BLIN_CTL_DUMP);
    if  (s >= 0) close(s);
    if  ((s = socket(PF_INET, SOCK_RAW, IPPROTO_DIVERT)) < 0) {
        ifBLIN_QW0("socket()");
        ex = EX_OSERR;
        goto out;
    }
    bzero(&sin, sizeof(sin));
    sin.sin_family = PF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(icfg->port);
    if  (bind(s, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
        ifBLIN_QW0("bind()");
        ex = EX_OSERR;
        goto out;
    }
    pidsave(icfg);
    sterm = shup = 0;
    if  (signal(SIGTERM, sigterm_handler) == SIG_ERR) ifBLIN_QW0("signal(TERM)");
    if  (signal(SIGINT, sigterm_handler) == SIG_ERR) ifBLIN_QW0("signal(INT)");
    if  (signal(SIGHUP, sighup_handler) == SIG_ERR) ifBLIN_QW0("signal(HUP)");
    if  ((kq = kqueue()) < 1) {
        ifBLIN_QW0("kqueue()");
        ex = EX_OSERR;
        goto out;
    }
    EV_SET(&kev, s, EVFILT_READ, EV_ADD, 0, 0, NULL);
    if  (!ipacket) {
        if  (!(ipacket = malloc(IP_MAXPACKET))) {
            ifBLIN_QX0("ipacket");
            ex = EX_OSERR;
            goto out;
    }   }
    if  (kevent(kq, &kev, 1, NULL, 0, NULL) < 0) {
        ifBLIN_QW0("kevent() set");
        ex = EX_OSERR;
        goto out;
    }
    while(1) {
        tcpshdr         *tcpacket;
        socklen_t        sinlen  ;
        int              optlen  ; /*    TCP  */
        ssize_t          plen    ; /*        */
        ssize_t          hlen    ; /*  IP                 */
        int              tlen    ; /*  TCP                */
        int              olen    ;
        int              mode    ; /*                 */
        u_char          *opt     ; /*   TCP        */
        int32_t          mss     ;
        ssize_t          i       ;

        tcpacket = NULL;
        switch(kevent(kq, NULL, 0, &kev, 1, NULL)) {
        default: if  (errno == EINTR) goto intrpt;
               ; ifBLIN_QW0("kevent() get");
               ; REEND(EX_OSERR);
               ; /* NOTREACHED */
        case 1 : mss = -1;
               ; sinlen = sizeof(sin);
               ; plen = recvfrom(s, ipacket, IP_MAXPACKET, 0, (struct sockaddr *)&sin, &sinlen);
               ; if  (plen < 0) {
               ;     if (errno == EINTR || errno == EAGAIN) continue;
               ;     ifBLIN_QW0("recvfrom");
               ;     REEND(EX_IOERR);
               ; }
               ; hlen = ipacket->ip_hl << 2;
               ; if  (  ((size_t)plen > sizeof(struct ip))
                     && (plen >= (hlen = ipacket->ip_hl << 2))
                     && (ipacket->ip_v == 4)
                     && (ipacket->ip_p == IPPROTO_TCP)
                     && !(ntohs(ipacket->ip_off) & IP_OFFMASK)
                     && (ntohs(ipacket->ip_len) == plen)
                     && ((plen - hlen) >= (ssize_t)sizeof(struct tcphdr))
                     ) {
               ;     mss = -2;
               ;     tcpacket = (tcpshdr *)((char *)ipacket + hlen); /* align == 2 */
               ;     tlen = tcpacket->th_off << 2;
               ;     if  (tlen > (plen - hlen)) {
               ;         ifBLIN_QX1("tlen %d > plen %zd - hlen %zd", tlen, plen, hlen);
               ;         mode = 2;
               ;         goto push;
               ;     }
               ;     mss = -3;
               ;     if  (!(tcpacket->th_flags & TH_SYN)) continue;
               ;     mss = -4;
               ;     if  (tlen > (ssize_t)sizeof(struct tcphdr)) {
               ;         for ( olen = tlen - (int)sizeof(struct tcphdr), opt = (u_char *)(tcpacket + 1)
                             ; olen > 0
                             ; olen -= optlen, opt += optlen
                             ) {
               ;             if  (*opt == TCPOPT_EOL) {
               ;                 mss = -5;
               ;                 break;
               ;             } else if (*opt == TCPOPT_NOP) {
               ;                 optlen = 1;
               ;             } else if ((optlen = *(opt + 1)) <= 0) {
               ;                 mss = -6;
               ;                 break;
               ;             } else if (optlen > olen) {
               ;                 mss = -7;
               ;                 break;
               ;             } else if (*opt == TCPOPT_MAXSEG) {
               ;                 mss = ntohs(recobe_load((opt + 2), 2));
               ;                 break;
               ;         }   }
               ;         mode = 0;
               ;     } else {
               ;         mode = 1;
               ;     }
               ; } else {
               ;     mode = 2;
               ; }
push:          ; for (i = icfg->top; i > 0; ) {
               ;     i--;
               ;     if  ((icfg->entry[i].min <= mss) && (icfg->entry[i].max >= mss)) {
               ;         if  (icfg->entry[i].rule >= 0) sin.sin_port = (in_port_t)icfg->entry[i].rule;
               ;         break;
               ; }   }
               ; ifBLIN_QO1 printlog(icfg, ipacket, tcpacket, mss, sin.sin_port, plen, mode);
               ; if  (sendto(s, ipacket, (size_t)plen, 0, (struct sockaddr *)&sin, sinlen) < 0) {
               ;     if (errno == EAGAIN) continue;
               ;     ifBLIN_QW0("sendto");
               ;     REEND(EX_IOERR);
               ; };
intrpt:        ; /* FALLTHROUGH */
        case 0 : if  (sterm) {
               ;     if  ((icfg->pidfi) && (0 > unlink(icfg->pidfi))) ifBLIN_QW0("unlink");
               ;     sterm = 0;
               ;     ifBLIN_QX0("TERM");
               ;     exit(0);
               ; }
               ; if  (shup) {
               ;     ifBLIN_QX0("HUP");
               ;     if  (!(icfg = mssseparated_readcfg(icfg))) {
               ;         ex = EX_OSERR;
               ;         REEND(EX_IOERR);
               ;     }
               ;     if  (icfg->logg) fclose(icfg->logg);
               ;     if  (icfg->logfi) {
               ;         if  (!(icfg->logg = fopen(icfg->logfi, "a"))) {
               ;             ifBLIN_QW0("fopen %s", icfg->logfi);
               ;         }
               ;         if  (setvbuf(icfg->logg, NULL, _IONBF, 0)) {
               ;             ifBLIN_QW0("setvbuf()");
               ;     }   }
               ;     pidsave(icfg);
               ;     shup = 0;
               ; }
               ; continue;
    }   }
out:
    if  (0 > mife_fini(icfg->md)) ifBLIN_QW1("mife_fini");
    exit(ex);
#   undef blin_internal_flags
}
